#include <WiFi.h>
#include <WebServer.h>
#include <Wire.h>
#include "MAX30105.h"
#include "heartRate.h"

/* Put your SSID & Password */
const char* ssid = "ESP32";  // Enter SSID here
const char* password = "12345678";  //Enter Password here

/* Put IP Address details */
IPAddress local_ip(192, 168, 1, 1);
IPAddress gateway(192, 168, 1, 1);
IPAddress subnet(255, 255, 255, 0);

WebServer server(80);

MAX30105 particleSensor;

const byte RATE_SIZE = 4; //Increase this for more averaging of IR values. 4 is good.
byte rates[RATE_SIZE]; //Array of heart rates
byte rateSpot = 0;
long lastBeat = 0; //Time at which the last beat occurred

float beatsPerMinute;
int beatAvg;
int display_ct = 0;

const int NUM_READINGS = 5; //Adjust the number of readings before displaying Heart Rate
int valid_HR[NUM_READINGS];
int index_valid = 0;
int display_HR;

int LED1pin = 2;
bool LED1status = LOW;
int LED2pin = 5;
bool LED2status = LOW;
int BUTTONpin = 10;
bool BUTTONstate = LOW;
int loopCounter = 0;
const int TOGGLE_INTERVAL = 50;  // for LEDs switching

enum State {
  INIT,
  MEASURE,
};

State currentState = INIT;
bool buttonPressed = false;
int lastAchievedHeartRate = 0;

int lastUpdateMillis = 0; // Variable to store the last update time
int lastnonValidMillis = 0; // Variable to store the last non valid time
int UPDATE_INTERVAL = 1000; // Update interval in milliseconds (adjust as needed)
int current_interval = 0; // Current interval between last valid reading
int noFinger_interval = 0; // Current interval between last not valid reading


// create function to check if there are 0 readings
bool hasZeroElement() {
  for (int i = 0; i < NUM_READINGS; i++) {
    if (valid_HR[i] == 0) {
      return true;
    }
  }
  return false;
}


void setup() {
  Serial.begin(115200);
  pinMode(LED1pin, OUTPUT);
  pinMode(LED2pin, OUTPUT);
  pinMode(BUTTONpin, INPUT);

  WiFi.softAP(ssid, password);
  WiFi.softAPConfig(local_ip, gateway, subnet);
  delay(100);

  //server set up
  server.on("/", handle_OnConnect);
  server.on("/start", HTTP_POST, handleStart);
  server.begin();
  Serial.println("HTTP server started");

  // Initialize sensor
  if (!particleSensor.begin(Wire, I2C_SPEED_FAST)) //Use default I2C port, 400kHz speed
  {
    Serial.println("MAX30105 was not found. Please check wiring/power. ");
    while (1);
  }

  particleSensor.setup(); //Configure sensor with default settings
  particleSensor.setPulseAmplitudeRed(0x0A); //Turn Red LED to low to indicate sensor is running

  // Initialize valid HR reading array
  for (int i = 0; i < NUM_READINGS; i++) {
    valid_HR[i] = 0;
  }

}
void loop() {
  server.handleClient();

  switch (currentState) {
    case INIT:
      digitalWrite(LED1pin, LOW);
      digitalWrite(LED2pin, LOW);
      if (digitalRead(BUTTONpin) == HIGH) {
        currentState = MEASURE;
        digitalWrite(LED1pin, HIGH);
        digitalWrite(LED2pin, HIGH);
        Serial.println("Pulse oximeter initialized.");
        Serial.println("When you're ready, place your index finger on the sensor with steady pressure and press the push button.");
      }
      delay(500);
      break;


    case MEASURE:
      long irValue = particleSensor.getIR();

      // Toggle between LED_PIN_1 and LED_PIN_2
      loopCounter++;
      if (loopCounter >= TOGGLE_INTERVAL) {
        static bool toggle = false;
        digitalWrite(toggle ? LED1pin : LED2pin, HIGH);
        digitalWrite(toggle ? LED2pin : LED1pin, LOW);
        toggle = !toggle;

        // Reset the loop counter
        loopCounter = 0;
      }

      if (checkForBeat(irValue) == true)
      {
        //We sensed a beat!
        long delta = millis() - lastBeat;
        lastBeat = millis();

        beatsPerMinute = 60 / (delta / 1000.0);

        if (beatsPerMinute < 220 && beatsPerMinute > 30)
        {
          rates[rateSpot++] = (byte)beatsPerMinute; //Store this reading in the array
          rateSpot %= RATE_SIZE; //Wrap variable

          //Take average of readings
          beatAvg = 0;
          for (byte x = 0 ; x < RATE_SIZE ; x++)
            beatAvg += rates[x];
          beatAvg /= RATE_SIZE;
        }
      }

      // Update the running average or non valid messages at regular intervals
      current_interval = millis() - lastUpdateMillis;
      noFinger_interval = millis() - lastnonValidMillis;

      if (irValue < 50000 && noFinger_interval >= 1000) {           // no valid reading
        lastnonValidMillis = millis();

        Serial.print(" No finger?");
        Serial.println();
      }

      else if (irValue >= 50000 && current_interval >= UPDATE_INTERVAL && beatAvg <125 && beatAvg >0) {  // sensing a presence, will determine if reading is valid

        lastUpdateMillis = millis();

        valid_HR[index_valid] = beatAvg;
        index_valid = (index_valid + 1) % NUM_READINGS;    // make sure you will only fill array until NUM_READINGS

        Serial.print("Valid BPMs are= ["); Serial.print(valid_HR[0] ); Serial.print("] [");
        Serial.print(valid_HR[1] ); Serial.print("] ["); Serial.print(valid_HR[2] ); Serial.print("] [");
        Serial.print(valid_HR[3] ); Serial.print("] ["); Serial.print(valid_HR[4] ); Serial.print("]");
       /* Serial.print(valid_HR[5] ); Serial.print("] ["); Serial.print(valid_HR[6] ); Serial.print("] [");
        Serial.print(valid_HR[7] ); Serial.print("] ["); Serial.print(valid_HR[8] ); Serial.print("] [");
        Serial.print(valid_HR[9] ); Serial.print("]");*/
        Serial.println();

        // Calculate the average and difference between max and min
        int sum = 0;
        int maxVal = valid_HR [0];
        int minVal = valid_HR [0];

        for (int i = 0; i < NUM_READINGS; i++) {
          sum += valid_HR[i];
          maxVal = max(valid_HR [i], maxVal);
          minVal = min(valid_HR [i], minVal);
        }
        int diffVal = maxVal - minVal;
        int avgHeartRate = sum / NUM_READINGS;

        if (diffVal <= 5) { // Adjust the threshold as needed

          Serial.print("\nYour current heart rate is ");
          Serial.print(avgHeartRate);
          Serial.print(" beats per minute.");

          currentState = INIT; // Transition to INIT state
          lastAchievedHeartRate = avgHeartRate;

          Serial.println("If you want to rescan, press either the web or physical button again.");
          // Clear the stored Heart Rate
          for (int i = 0; i < NUM_READINGS; i++) {
            valid_HR[i] = 0;
          }
        } else if (hasZeroElement()) {
          Serial.println("Initializing your reading");
        } else if (avgHeartRate <= 40) {
          Serial.println("You might not be making enough contact. Apply more pressure");
        } else {
          Serial.println("Calculating your heart rate...");
        }
      }
  }
}


void handle_OnConnect() {
  String buttonLabel = (currentState == INIT) ? "Start Measurement" : "Stop Measurement";
  String html = "<html><head><style> button {font-size: 18px;";
  html += " padding: 10px 20px; } ";
  html += "html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}\n</style> ";
  html += "<meta http-equiv='refresh' content='2'>"; // Refresh the page every 5 seconds (adjust as needed)
  html += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, user-scalable=no\">\n";
  html += "</head><body>";
  html += "<h1>ESP32 Pulse Oximeter</h1>";
  html += "<p>Current State: ";
  html += (currentState == INIT) ? "INIT" : "MEASURE";
  html += "</p>";
  html += "<form action=\"/start\" method=\"post\">";
  html += "<button type=\"submit\">" + buttonLabel + "</button>";
  html += "</form>";
  html += "<p>If you want to scan, place your index finger on the sensor with steady pressure and press either the web or physical button.</p>";
  html += "<p></p><p>Last Achieved Heart Rate is : <b>" + String(lastAchievedHeartRate) + " BPM </b></p>";
  html += "</body></html>";
  server.send(200, "text/html", html);
}

void handleStart() {
  if (currentState == INIT) {
    currentState = MEASURE;
    buttonPressed = true;
    server.sendHeader("Location", "/");
  } else {
    currentState = INIT;
    buttonPressed = false;
    lastAchievedHeartRate = 0; // Reset last achieved heart rate when stopping measurement
    server.sendHeader("Location", "/");
  }
  server.send(302, "text/plain", "");
}
